/*
 * 2ooM: The Master of Orion II Reverse Engineering Project
 * Library for working with LBX image files.
 * Copyright © 2006-2011, 2013-2014 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <inttypes.h>
#include <errno.h>

#include "pack.h"
#include "misc.h"
#include "lbx.h"
#include "error.h"
#include "image.h"

#define FLAG_RAW       0x0100 /* Image is stored as a flat array of bytes. */
#define FLAG_OVERWRITE 0x0400 /* Draw each frame on a clean slate (unsure). */
#define FLAG_BUILDING  0x0800 /* Buildings have this, related to shadow? */
#define FLAG_PALETTE   0x1000 /* Image contains embedded palette. */
#define FLAG_LOOPING   0x2000 /* Loop over all frames in the image (unsure). */

#define FLAG_ALL (FLAG_RAW|FLAG_OVERWRITE|FLAG_BUILDING|FLAG_PALETTE|FLAG_LOOPING)

#define HDR_LEN 12

struct lbx_image_priv {
	struct lbx_image pub;

	unsigned short wtf, flags;
	unsigned char  wtf2;
	unsigned short palstart, palcount;

	const struct lbx_file_ops *fops;
	int (*dtor)(void *handle);
	void *f;

	long paloff;

	int currentframe;
	int currentx, currenty;
	unsigned char **framedata;
	unsigned char **mask;

	unsigned long offsets[];
};

static struct lbx_image_priv *lbx_img_init(unsigned char hdr[static HDR_LEN])
{
	unsigned short nframes = unpack_16_le(hdr+6);
	struct lbx_image_priv *img;

	img = malloc(sizeof *img + sizeof img->offsets[0] * (nframes+1));
	if (!img) {
		lbx_error_raise(LBX_ENOMEM);
		return NULL;
	}

	*img = (struct lbx_image_priv) {
		.pub.width  = unpack_16_le(hdr+0),
		.pub.height = unpack_16_le(hdr+2),
		.wtf        = unpack_16_le(hdr+4),
		.pub.frames = hdr[6],
		.wtf2       = hdr[7],
		.pub.leadin = hdr[8],
		.pub.chunk  = hdr[9],
		.flags      = unpack_16_le(hdr+10),

		.currentframe = -1,
	};

	if (img->flags & FLAG_OVERWRITE)
		img->pub.chunk = 1;

	if (img->flags & FLAG_LOOPING)
		img->pub.leadin = 0;

	return img;
}

struct lbx_image *lbx_img_open(void *f, const struct lbx_file_ops *fops,
                               int (*destructor)(void *))
{
	unsigned char hdr_buf[HDR_LEN];
	struct lbx_image_priv *img;

	if (fops->read(hdr_buf, sizeof hdr_buf, f) != sizeof hdr_buf) {
		if (fops->eof(f))
			lbx_error_raise(LBX_EEOF);
		return NULL;
	}

	img = lbx_img_init(hdr_buf);
	if (!img)
		return NULL;

	img->f    = f;
	img->fops = fops;
	img->dtor = destructor;

	/*
	 * DEBUG ONLY.  These assertions exist to catch otherwise valid image
	 * files which differ from what I believe to be true of all LBX images.
	 * When we can decode every image, then these assertions should be
	 * replaced with constraints.
	 */
	_lbx_assert(img->wtf  == 0); /* version? */
	_lbx_assert(img->wtf2 == 0); /* very likely is simply reserved. */
	_lbx_assert(img->pub.frames > img->pub.leadin);
	_lbx_assert(!(img->flags & ~FLAG_ALL));

	/* Read all offsets.  Should be merged with identical code in lbx.c */
	for (unsigned i = 0; i <= img->pub.frames; i++) {
		unsigned char buf[4];

		if (fops->read(buf, sizeof buf, f) != sizeof buf) {
			if (fops->eof(f))
				lbx_error_raise(LBX_EEOF);
			free(img);
			return NULL;
		}

		img->offsets[i] = unpack_32_le(buf);
	}

	if (img->flags & FLAG_PALETTE) {
		unsigned char buf[4];

		if (fops->read(buf, sizeof buf, f) != sizeof buf) {
			if (fops->eof(f))
				lbx_error_raise(LBX_EEOF);
			free(img);
			return NULL;
		}

		img->palstart = unpack_16_le(buf+0);
		img->palcount = unpack_16_le(buf+2);
		img->paloff   = fops->tell(f);

		if (img->palstart + img->palcount > 256) {
			lbx_error_raise(LBX_EFORMAT);
			free(img);
			return NULL;
		}
	}

	return &img->pub;
}

static int pipe_close(void *f)
{
	struct lbx_pipe_state *p = f;
	int rc;

	rc = fclose(p->f);
	free(p);
	return rc;
}

static int file_close(void *f)
{
	return fclose((FILE *)f);
}

struct lbx_image *lbx_img_fopen(const char *file)
{
	struct lbx_pipe_state *p;
	FILE *f;

	f = fopen(file, "rb");
	if (!f) {
		lbx_error_raise(-errno);
		return NULL;
	}

	if (fseek(f, 0, SEEK_CUR) == 0)
		return lbx_img_open(f, &lbx_default_fops, file_close);

	p = malloc(sizeof *p);
	if (!p) {
		lbx_error_raise(LBX_ENOMEM);
		fclose(f);
		return NULL;
	}

	*p = (struct lbx_pipe_state) { .f = f };
	return lbx_img_open(p, &lbx_pipe_fops, pipe_close);
}

static int _lbx_drawrow(struct lbx_image_priv *img)
{
	unsigned short length, offset;
	unsigned char buf[4];
	unsigned char *pos;
	size_t rc;

	assert(img->framedata);
	assert(img->mask);

	if (img->fops->read(buf, sizeof buf, img->f) != sizeof buf)
		goto readerr;

	length = unpack_16_le(buf+0);
	offset = unpack_16_le(buf+2);
	if (length == 0 && offset == 1000)
		return 1;

	/* Length of 0 increments Y position */
	if (!length) {
		if (offset > img->pub.height - img->currenty) {
			lbx_error_raise(LBX_EFORMAT);
			return -1;
		}

		img->currenty += offset;
		img->currentx  = 0;
		return 0;
	}

	/* Otherwise we read pixel data */
	if (offset > img->pub.width - img->currentx) {
		lbx_error_raise(LBX_EFORMAT);
		return -1;
	}
	img->currentx += offset;

	if (length > img->pub.width - img->currentx) {
		lbx_error_raise(LBX_EFORMAT);
		return -1;
	}

	memset(&img->mask[img->currenty][img->currentx], 1, length);

	pos = &img->framedata[img->currenty][img->currentx];
	rc  = img->fops->read(pos, length, img->f);
	img->currentx += rc;

	if (rc < length)
		goto readerr;

	/* Skip padding byte, if any */
	if (length % 2) {
		if (img->fops->read(buf, 1, img->f) != 1)
			goto readerr;
	}

	return 0;
readerr:
	if (img->fops->eof(img->f))
		lbx_error_raise(LBX_EEOF);
	return -1;
}

static unsigned char **allocframebuffer(size_t width, size_t height)
{
	unsigned char **new, *tmp;
	size_t i;

	if (height > SIZE_MAX / sizeof *new) {
		lbx_error_raise(LBX_ENOMEM);
		return NULL;
	}

	/* Ensure that there is at least one row in the framebuffer. */
	if (height == 0 || width == 0)
		width = height = 1;

	tmp = calloc(height, width);
	if (!tmp) {
		lbx_error_raise(LBX_ENOMEM);
		return NULL;
	}

	new = malloc(height * sizeof *new);
	if (!new) {
		lbx_error_raise(LBX_ENOMEM);
		free(tmp);
		return NULL;
	}

	for (i = 0; i < height; i++) {
		new[i] = tmp + i * width;
	}

	return new;
}

static unsigned char **read_raw_frame(struct lbx_image_priv *img, int frame)
{
	unsigned long size = img->pub.width * img->pub.height;

	assert(img->flags & FLAG_RAW);

	if (img->fops->seek(img->f, img->offsets[frame], SEEK_SET)) {
		return NULL;
	}

	if (img->fops->read(img->framedata[0], size, img->f) != size) {
		if (img->fops->eof(img->f))
			lbx_error_raise(LBX_EEOF);
		return NULL;
	}
	memset(img->mask[0], 1, size);

	if (img->fops->tell(img->f) > img->offsets[frame+1]) {
		lbx_error_raise(LBX_EFORMAT);
		return NULL;
	}

	return img->framedata;
}

unsigned char **lbx_img_getframe(struct lbx_image *pub, int frame)
{
	struct lbx_image_priv *img = (struct lbx_image_priv *)pub;
	unsigned char buf[4];

	if (frame >= pub->frames || frame < 0) {
		lbx_error_raise(LBX_EINVAL);
		return NULL;
	}

	if (!img->framedata) {
		img->framedata = allocframebuffer(pub->width, pub->height);
		if (!img->framedata)
			return NULL;
	}

	if (!img->mask) {
		img->mask = allocframebuffer(pub->width, pub->height);
		if (!img->mask)
			return NULL;
	}

	if (img->flags & FLAG_RAW)
		return read_raw_frame(img, frame);

	if ((img->flags & FLAG_OVERWRITE)
	     || (pub->chunk && !(frame % pub->chunk))) {
		/* Clear the slate. */
		img->currentframe = -1;
		memset(img->framedata[0], 0, pub->width * pub->height);
		memset(img->mask[0],      0, pub->width * pub->height);
	} else {
		/* Start over if we are backtracking. */
		if (img->currentframe > frame) {
			memset(img->mask[0], 0, pub->width * pub->height);
			img->currentframe = -1;
		}

		/* We must have previous frame decoded to continue. */
		if (frame > img->currentframe + 1) {
			if (!lbx_img_getframe(pub, frame-1))
				return NULL;
		}
	}

	if (img->currentframe != frame) {
		int rc;

		if (img->fops->seek(img->f, img->offsets[frame], SEEK_SET)) {
			return NULL;
		}

		/* Read frame header */
		if (img->fops->read(buf, 4, img->f) != 4) {
			if (img->fops->eof(img->f))
				lbx_error_raise(LBX_EEOF);
			return NULL;
		}

		if (unpack_16_le(buf) != 1) {
			lbx_error_raise(LBX_EFORMAT);
			return NULL;
		}

		img->currentx = 0;
		img->currenty = unpack_16_le(buf+2);
		if (img->currenty > img->pub.height) {
			lbx_error_raise(LBX_EFORMAT);
			return NULL;
		}

		do {
			rc = _lbx_drawrow(img);
			if (rc == -1)
				return NULL;

			if (img->fops->tell(img->f) > img->offsets[frame+1]) {
				lbx_error_raise(LBX_EFORMAT);
				return NULL;
			}
		} while (!rc);
	}

	img->currentframe = frame;
	return img->framedata;
}

int
lbx_img_loadpalette(void *f, const struct lbx_file_ops *fops,
                    struct lbx_colour palette[static 256])
{
	unsigned char entry[4];
	int i;

	for (i = 0; i < 256; i++) {
		if (fops->read(entry, sizeof entry, f) != sizeof entry) {
			if (fops->eof(f))
				lbx_error_raise(LBX_EEOF);
			return -1;
		}

		if (entry[0] != 1) {
			lbx_error_raise(LBX_EFORMAT);
			return -1;
		}

		palette[i] = (struct lbx_colour) {
			.red    = entry[1] & 0x3f,
			.green  = entry[2] & 0x3f,
			.blue   = entry[3] & 0x3f,
			.active = 1,
		};
	}

	return 0;
}

int
lbx_img_getpalette(struct lbx_image *pub, struct lbx_colour palette[static 256])
{
	struct lbx_image_priv *img = (struct lbx_image_priv *)pub;
	unsigned char entry[4];
	unsigned int i;
	size_t rc;

	/* Do nothing if the image doesn't have embedded palette data. */
	if (!(img->flags & FLAG_PALETTE))
		return 0;

	if (img->fops->seek(img->f, img->paloff, SEEK_SET)) {
		return -1;
	}

	for (i = 0; i < img->palcount; i++) {
		rc = img->fops->read(entry, sizeof entry, img->f);
		if (rc < sizeof entry) {
			goto readerr;
		}

		if (entry[0] != 0) {
			lbx_error_raise(LBX_EFORMAT);
			return -1;
		}

		palette[img->palstart + i] = (struct lbx_colour){
			.red    = entry[1],
			.green  = entry[2],
			.blue   = entry[3],
			.active = 1,
		};
	}

	return 0;
readerr:
	if (img->fops->eof(img->f))
		lbx_error_raise(LBX_EEOF);
	return -1;
}

void lbx_img_getinfo(struct lbx_image *pub, struct lbx_imginfo *info)
{
	struct lbx_image_priv *img = (struct lbx_image_priv *)pub;

	*info = (struct lbx_imginfo) {
		.palettesz  = (img->flags & FLAG_PALETTE) ? img->palcount : 0,
	};

	/* There seems to be two ways of specifying that an image loops. */
	if (img->flags & FLAG_LOOPING) {
		info->loopstart = 0;
		info->looping   = 1;
	} else if (img->pub.leadin != pub->frames - 1) {
		info->loopstart = img->pub.leadin;
		info->looping   = 1;
	}
}

unsigned char **lbx_img_getmask(struct lbx_image *pub)
{
	struct lbx_image_priv *img = (struct lbx_image_priv *)pub;

	return img->mask;
}

int lbx_img_close(struct lbx_image *pub)
{
	struct lbx_image_priv *img = (struct lbx_image_priv *)pub;
	int rc = 0;

	if (img) {
		if (img->framedata) {
			free(img->framedata[0]);
			free(img->framedata);
		}

		if (img->mask) {
			free(img->mask[0]);
			free(img->mask);
		}

		if (img && img->dtor) {
			rc = img->dtor(img->f);
		}

		free(img);
	}

	return rc;
}
